//
//  GMlib -- Graphics & Media Lab Common Source Library
//
//  $Id: gmlpathstring.cpp,v 1.12 2004/11/06 17:56:59 04a_deg Exp $
//
//  Copyright (C) 2004, Moscow State University Graphics & Media Lab
//  gmlsupport@graphics.cs.msu.su
//  
//  This file is part of GMlib software.
//  For conditions of distribution and use, see the accompanying README file.

#include "../base/gmlcommon.h"
#include "gmlpathstring.h"

#include <vector>
#include <algorithm>

#ifndef _WIN32_WCE
#include <sys/stat.h>
#include <direct.h>
#endif

#include <windows.h>

using namespace gml;
using std::vector;

const char gml::PathString::PATH_SEPARATOR = '\\';
const string gml::PathString::PATH_SEPARATOR_STR = string("\\");
const int MAX_PATH_LEN = 500;

PathString::PathString(const string& path,
                       const string& filename,
                       const string& extension)
{
  ASSERT(!filename.empty());

  *this = "";
  *this += path;

  int len = length();
  if (len > 0 && (*this)[len - 1] != PATH_SEPARATOR)
  {
    *this += PATH_SEPARATOR_STR;
  }

  *this += filename;

  if (!extension.empty())
  {
    if (extension[0] != '.')
      *this += ".";
    *this += extension;
  }

  CorrectSeparators();
}


string PathString::Path() const
{
  int pos = find_last_of(PATH_SEPARATOR);
  if (pos == npos)
    return string();

  // Check for logical drive presence
  if (pos > 0 && (*this)[pos - 1] == ':')
    pos++;

  // Check for root path
  if (pos == 0)
    pos++;

  return substr(0, pos);
}

string PathString::Filename() const
{
  int pos = find_last_of(PATH_SEPARATOR);
  if (pos == npos)
    return *this;

  ++pos;
  return substr(pos);
}

string PathString::Basename() const
{
  string fname = Filename();
  int pos = fname.find_last_of('.');
  if (pos == npos)
    return fname;

  return fname.substr(0, pos);
}

string PathString::Extension() const
{
  string fname = Filename();
  int pos = fname.find_last_of('.');
  if (pos == npos)
    return string();

  return fname.substr(pos + 1);
}

void PathString::Split(string* path, string* filename, string* extension) const
{
  if (path != NULL)
    *path = (*this).Path();

  if (filename != NULL)
  {
    if (extension != NULL)
      *filename = (*this).Basename();
    else
      *filename = (*this).Filename();
  }

  if (extension != NULL)
    *extension = (*this).Extension();
}

#if __BORLANDC__ >= 0x550  || !defined(__BORLANDC__)

bool PathString::IsFile() const
{
#ifndef _WIN32_WCE
  struct _stat s;
  int res = _stat(c_str(), &s);
  return res == 0 && (s.st_mode & _S_IFREG);
#else
  WIN32_FIND_DATA findFileData;
  HANDLE hFind = FindFirstFile(gml::String::ToUnicode(*this).c_str(), &findFileData);
	if (hFind == INVALID_HANDLE_VALUE)
		return FALSE;
	VERIFY(FindClose(hFind));

	return findFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY == 0;

#endif
}

bool PathString::IsDir() const
{
#ifndef _WIN32_WCE
  struct _stat s;
  int res = _stat(c_str(), &s);
  return res == 0 && (s.st_mode & _S_IFDIR);
#else
  WIN32_FIND_DATA findFileData;
  HANDLE hFind = FindFirstFile(gml::String::ToUnicode(*this).c_str(), &findFileData);
	if (hFind == INVALID_HANDLE_VALUE)
		return FALSE;
	VERIFY(FindClose(hFind));

	return findFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY != 0;

#endif
}


#ifndef _WIN32_WCE
bool PathString::CurDir()
{
  char buffer[255];

  // Receive CWD by getcwd() 
  if (_getcwd(buffer, 255) == NULL)
    return false;

  *this = buffer;
  return true;
}
#endif
#endif

#ifndef _WIN32_WCE
bool PathString::ChDir() const
{
  if (empty())
    return false;

  if (chdir(c_str()) != 0)
    return false;

  return true;
}
#endif

bool PathString::ExePath()
{
#ifndef _WIN32_WCE
  char buffer[500];
#else
  wchar_t buffer[500];
#endif
  if (GetModuleFileName(NULL, buffer, 500) == 0)
    return false;
#ifndef _WIN32_WCE
  *this = PathString(buffer);
#else
  *this = PathString(gml::String::FromUnicode(buffer));
#endif
  return true;
};



bool PathString::ListFileName(list<PathString>& fnames, PathString* mask) const
{
  //ASSERT(isdir());

  fnames.resize(0);
  // ------------------------------------------------
  //             Open catalog
  // ------------------------------------------------

  HANDLE handle = NULL;

  // ------------------------------------------------
  //             Read catalog
  // ------------------------------------------------

  int n_names = 0;

  WIN32_FIND_DATA fdata;
  PathString path(*this, "*.*");  // All items of the catalog
  // for the anologous with UNIX
  // Read first item from catalog list
#ifndef _WIN32_WCE
  handle = FindFirstFile(path.c_str(), &fdata);
#else
  handle = FindFirstFile(gml::String::ToUnicode(path).c_str(), &fdata);
#endif
  if (handle == INVALID_HANDLE_VALUE)
    return false;

  while (FindNextFile(handle, &fdata))
  {
    // Skip parent and current catalogs
    if ((fdata.cFileName[0] == '.' && fdata.cFileName[1] == '\0') ||
        (fdata.cFileName[0] == '.' &&
         fdata.cFileName[1] == '.' &&
         fdata.cFileName[2] == '\0'))
      continue;

#ifndef _WIN32_WCE
     PathString filename = PathString(fdata.cFileName);
#else
   
    PathString filename = PathString(gml::String::FromUnicode(fdata.cFileName));

#endif
    if (mask != NULL && filename.CheckMask(*mask) != true)
      continue;

    if (n_names == 0)
      (void) fnames.resize(0);

    fnames.push_back(filename);

    n_names++;
  }

  // ------------------------------------------------
  //             Close catalog
  // ------------------------------------------------


  if (!FindClose(handle))
    return false;

  return true;
}


////////////////////////////////////////////////////////////////////////////////////////
///
bool PathString::MkDir() const
{
  if (this->empty())
    return true;

#ifndef _WIN32_WCE
  if (mkdir(this->c_str()) != 0)
    return false;
#else
  if (!CreateDirectory(gml::String::ToUnicode(*this).c_str(), NULL))
    return false;
#endif

  return true;
}

////////////////////////////////////////////////////////////////////////////////////////
///
bool PathString::MakePath() const
{
  if (empty())
    return false;


  // make full path
  gml::PathString full_path(*this);
#ifndef _WIN32_WCE
  full_path.MakeFullPath();
#endif

  gml::PathString build_path;
  int previous_separator = full_path.find_first_of(PATH_SEPARATOR, 0);
  int next_separator = full_path.find_first_of(PATH_SEPARATOR,
                                               previous_separator + 1);
  build_path.assign(*this, 0, previous_separator);
  // skip first!
  for (;
       next_separator != npos;
       previous_separator = next_separator, next_separator = full_path.find_first_of(PATH_SEPARATOR, next_separator + 1))
  {
    build_path += PATH_SEPARATOR;

    gml::String dir_part(full_path,
                         previous_separator + 1,
                         next_separator - previous_separator - 1);
    build_path += dir_part.c_str();

    if (!build_path.IsDir())
    {
      bool res = build_path.MkDir();
      if (!res)
        return false;
    }
  }
  return true;
}

//////////////////////////////////////////////////////////////////////////////
///
bool PathString::RmDir() const
{
  if (this->empty())
    return false;

#ifndef _WIN32_WCE
  if (rmdir(this->c_str()) == 0)
    return true;
#else
  if (RemoveDirectory(gml::String::ToUnicode(*this).c_str()))
    return true;
#endif

  return false;
}

//////////////////////////////////////////////////////////////////////////////
///
bool PathString::RmFile() const
  {
  if (this->empty())
    return false;

#ifndef _WIN32_WCE
  if (unlink(this->c_str()) == 0)
    return true;
#else
  if (DeleteFile(gml::String::ToUnicode(*this).c_str()))
    return true;
#endif

  return false;
  }

int RecMatch(unsigned char* p, // sh pattern to match 
unsigned char* s)  // string to match it to
{
  unsigned int c;  // pattern char or start of range in [-] loop  

  // Get first character, the pattern for new recmatch calls follows 
  c = *p++;

  // If that was the end of the pattern, match if string empty too 
  if (c == 0)
    return *s == 0;

  // '?' matches any character (but not an empty string) 
  if (c == '?')
    return *s ? RecMatch(p, s + 1) : 0;

  // '*' matches any number of characters, including zero 
  if (c == '*')
  {
    if (*p == 0)
      return 1;
    for (; *s; s++)
      if ((c = RecMatch(p, s)) != 0)
        return c;
    return 2;           // 2 means give up--shmatch will return false 
  }

  // If escape ('\'), just compare next character 
  if (c == '\\')
    if ((c = *p++) == 0)    // if \ at end, then syntax error
      return 0;

  // Just a character--compare it 
#ifdef _WIN32
  if (toupper((int) c) == toupper(*s))
    #else                 // UNIX
    if (c == (*s))
                    #endif
      return RecMatch(p, ++s);

  return 0;
}


bool PathString::CheckMask(const PathString& mask) const
{
  const char* p1 = this->c_str();
  const char* p2 = mask.c_str();

  if (RecMatch((unsigned char *) p2, (unsigned char *) p1) != 1)
    return false;

  return true;
}

#ifndef _WIN32_WCE
#if __BORLANDC__ >= 0x550 || !defined(__BORLANDC__)
bool PathString::MakeFullPath(const PathString& relative_to)
{
  gml::PathString cur_path;
  if (!relative_to.empty())
  {
    // save current directory
    cur_path.CurDir();
    // and setup new dir
    bool res = relative_to.ChDir();

    //gml::PathString new_path(relative_to);
    //bool res = new_path.ChDir();

    if (!res)
    {
      cur_path.ChDir(); 
      return false;
    }
  }

  char full_path[MAX_PATH_LEN + 1];
  _fullpath(full_path, this->c_str(), MAX_PATH_LEN);

  *this = full_path;

  if (!relative_to.empty())
  {
    cur_path.ChDir();
  };

  return true;
}
#endif
#endif

void PathString::CorrectSeparators()
{
  std::replace(begin(), end(), '/', PATH_SEPARATOR);
}